/******************************************************************************
 *
 * LF.net ipcount $Id$
 *
 * ipcount snifft auf einem Netzwerk Interface die Pakete, und sortiert die
 * Anzahl der Pakete und deren Groessen-Summe in eine Matrix mit src:dst 
 * als Key ein. ipcount wird zu Accounting Zwecken verwendet.
 *
 * 
 * mah, Thu Jul 23 09:58:15 CEST 2009
 * - initiale Version, Rewrite der Version 3
 *
 ******************************************************************************/

#include <iostream>
#include <boost/program_options.hpp>
// apt-get install libboost-program-options-dev

#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <grp.h>
#include <pwd.h>

#include "pcap_iface.h"
#include "ipcount_error.h"
#include "ipv4.h"

using namespace std;
namespace po = boost::program_options;

// solange Pakete gezaehlt werden sollen, ist capturing true
volatile bool capturing = true;

// wird auf true gesetzt, wenn der stuendliche Traffic- und Stats-Dump
// noetig ist, und nach dem Dump wieder auf false gesetzt
volatile bool trafficdump_time = false;
volatile bool monitor_time = false;

// wird ein HUP Signal empfangen, so wird sofort ein Traffic- und
// Stats-Dump geschrieben
volatile bool hupped = false;

// loggt einen String in /var/log/daemons
void log2syslog(const string &s)
{
	// LOG_CONS: Write  directly  to  system  console  if there is an error
	// while sending to system logger.
	openlog("ipcount", LOG_CONS | LOG_PID | LOG_PERROR, LOG_DAEMON);
	syslog(LOG_INFO,"%s",s.c_str());
}

// generiert den File Timestamp mit optionalen Stunden-Offset. Das Offset
// wird verwendet, da zur vollen Stunde gedumped/rotiert wird und die
// Traffic und Dump Files aber die vorige Stunde im Namen haben muessen.
string output_timestamp(int houroffset)
{
	char s[128];
	time_t now = time(0);
	now += houroffset*3600;
	struct tm *tm = localtime(&now);
	strftime(s, sizeof(s)-1, "%Y%m%d%H0000", tm);
	return s;
}

// liefert einen Filenamen, bestehend aus versch. Teilen
//
//  dir = "/var/log"
//  prefix = "traffic"
//  offset = -1
//  output_filename -> "/var/log/traffic.20090721130000", wobei von der
//  Stunde eine abgezogen ist.
string output_filename(const string &dir, const string &prefix, int houroffset)
{
	stringstream ss;
	ss << dir << "/" << prefix << "." << output_timestamp(houroffset);
	return ss.str();
}

// der Signal Handler wird bei allen Signalen (asynchron) aufgerufen,
// fuer die in der Prozedur install_signal_handlers() ein Handler installiert
// wurde. Alle Variablen, die mit der "Aussenwelt" ausgetauscht werden muessen,
// sollten als volatile deklariert sein.

void signal_handler(int signal)
{
	switch (signal)
	{
		case SIGALRM :
			{
				// jede Sekunde kommt ein SIGALRM; jetzt schauen, ob es
				// Zeit fuer einen stuendlichen Dump ist:
				const time_t now = time(0);
				struct tm tm;

				// statische Zeitvar des letzten Laufs initialisieren:
				// die 3 Nullen setzen Sec,Min und Stunden
				// siehe /usr/include/time.h:134
				static struct tm last_tm = { 0, 0, 0 };
			   	localtime_r(&now, &tm);

				// Stundenueberlauf pruefen: wenn der Minutenwert des
				// letzten Males groesser ist als die jetztige Minute,
				// dann hat ein Minutenueberlauf (= naechste Stunde) 
				// stattgefunden. Diese Methode ist robust gegenueber der
				// Situation, dass die Minute und Sekunde 0 nicht 
				// "auftauchen", sprich zu diesem Zeitpunkt der SIGALRM
				// nicht ausgeloest wird.
				if (last_tm.tm_min > tm.tm_min)
				{
					trafficdump_time = true;
				}

				// bzw. es Zeit fuer einen monitoring.dump ist; auch hier
				// gleiches Verfahren wie beim Stundenueberlauf, nur als
				// 5-Minuten-Ueberlauf
				if ((last_tm.tm_min % 5) > (tm.tm_min % 5))
				{
					monitor_time = true;
				}

				// letzten Zeitpunkt merken
				last_tm = tm;

				// Timer wieder stellen
				alarm(1);
			}
			break;
		case SIGHUP :
			// wird in der Main Loop ausgewertet und sofort ein Dump erzeugt
			hupped = true;
			break;
		case SIGINT :
		case SIGTERM :
			// die Main Loop wird darauf hin verlassen und vor dem Beenden
			// noch gedumpt
			capturing = false;
			break;
	}
}

// installiert fuer diverse Signale die signal_handler(..) Routine
void install_signal_handlers()
{
	signal(SIGHUP, signal_handler);
	signal(SIGALRM, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGINT, signal_handler);
}

// prueft, ob ipcount auch als root gestartet wurde. Das ist noetig, um
// das Netzwerk Interface zu oeffnen
void check_uid()
{
	if (geteuid() != 0)
		throw ipcount_error("root permissions required");
}

// gibt die root Rechte ab
void drop_root_privs(const string &user, const string &group,
		const string &dir)
{
	// uid des Users nachschauen
	struct passwd *pw = getpwnam(user.c_str());
	if (! pw)
		throw ipcount_error("unknown user: " + user);

	// gid der Gruppe nachschauen
	struct group *gr = getgrnam(group.c_str());
	if (! gr)
		throw ipcount_error("unknown group: " + group);

	// zuletzt User/Gruppe wechseln

	if (setgid(gr->gr_gid) != 0)
		throw ipcount_error("setgid error");
	if (setegid(gr->gr_gid) != 0)
		throw ipcount_error("setegid error");

	if (setuid(pw->pw_uid) != 0)
		throw ipcount_error("setuid error");
	if (seteuid(pw->pw_uid) != 0)
		throw ipcount_error("seteuid error");
}

// prueft, ob im Output Dir auch geschrieben werden kann
void check_write_perms(const string &dir)
{
	const string filename = dir + "/ipcount-write-test";
	FILE *f = fopen(filename.c_str(),"w");
	if (f)
	{
		// falls erfolgreich, dann schliessen und gleich wieder loeschen
		fclose(f);
		if (unlink(filename.c_str()) != 0)
			throw ipcount_error("could not unlink '" + filename + 
					"': " + strerror(errno));
	}
	else
		throw ipcount_error("could not write to '" + dir + 
				"': " + strerror(errno));
}

// schreibt ein file mit der PID des Prozesses
void write_pid_file(const string &interface)
{
	const string filename = "/var/run/ipcount-" + interface + ".pid";
	FILE *f = fopen(filename.c_str(),"w");
	if (f)
	{
		fprintf(f,"%u\n",getpid());
		fclose(f);
	}
	else
		throw ipcount_error("could not write '" + filename + 
				"': " + strerror(errno));
}

int main(int argc, char *argv[])
{
	// Globaler Exception try/catch Block; es wird fuer eigene Fehler
	// die ipcount_error Exception Klasse verwendet. Aber auch die anderen
	// Exceptions wie von der Boost-Lib und der Std-Lib werden gefangen.
	try
	{

		// Hier wird die Program Options Lib auf der Boost Library verwendet.
		// Vorteil ist das bequeme Parsen, unter Lenny ist die Boost-Lib
		// zu installieren mit:
		//
		// (devel): apt-get install libboost-program-options-dev
		// (run):   get-get install libboost-program-options1.34.1
		//
        // http://www.boost.org/doc/libs/1_34_1/doc/html/program_options.html
		//
		po::options_description options("ipcount options");
		options.add_options()
			("help,h", "shows the available options")
			("interface,i", po::value<string>(), 
			 	"network interface to account on")
			("outdir,o", po::value<string>(), "directory to write output to")
			("user,u", po::value<string>(), "username, default: nobody")
			("group,g", po::value<string>(), "groupname, default: nogroup")
			;

		// Options verarbeiten:
		po::variables_map map;
		po::parsed_options p_options = po::parse_command_line(argc, 
				argv, options);
		po::store(p_options, map);
		po::notify(map);

		// Pruefen, ob alles korrekt angegeben wurde:
		if (map.count("help") || 
				! map.count("interface") ||
				! map.count("outdir"))
		{
			cerr << "LF.net ipcount 4.0 - IP volume accouting tool" 
				<< endl << endl;
			cerr << options << endl;
			cerr << 
				"Usage: ipcount -i <interface> -o <dir> [-u user -g group -c]"
				<< endl;
			exit(1);
		}

		string outputdir = map["outdir"].as<string>();

		// Checken, ob wir root sind
		check_uid();

		// User auswerten, default ist "nobody"
		string username = "nobody";
		if (map.count("user"))
			username = map["user"].as<string>();

		// Gruppe auswerten, default ist "nogroup"
		string groupname = "nogroup";
		if (map.count("group"))
			groupname = map["group"].as<string>();

		// Signal-Handler installieren
		install_signal_handlers();

		// pid File schreiben:
		write_pid_file(map["interface"].as<string>());

		// PCAP Interface oeffnen; dabei wird auch das net device
		// geoeffnet und in den Promisc Mode gesetzt
		pcap_iface pcap(map["interface"].as<string>());

		// Filter setzen; bei IPv6 muss hier vermutlich etwas geaendert werden
		pcap.set_filter("ip");

		// root Rechte abgeben
		drop_root_privs(username, groupname, outputdir);

		// Dir Check - ist Schreiben moeglich?
		check_write_perms(outputdir);

		// Ticker starten, wird jede Sekunde aufgerufen; wird verwendet, um
		// zur vollen Stunde zu dumpen
		alarm(1);

		log2syslog("ipcount started capturing on " + 
				map["interface"].as<string>());

		// Capture Schleife:
		while (1)
		{
			// Paket von der libpcap abholen lassen; der Call ist blocking

			pcap.capture_packet(ipv4_packet_handler);

			if (monitor_time)
			{
				monitor_time = false;

				// Pointer auf die Error-Counter struct holen
				pcap_counters *counters = pcap.counters();

				// Drops seit dem letzten Stats-Messzeitpunkt besorgen
				uint64_t deltadrops = pcap.get_delta_drops();
				// und zu den capture_drop addieren
				counters->capture_drop += deltadrops;
				counters->deltadrops = deltadrops;

				// Filename fuer Monitor-Dump zusammensetzen
				const string filename = outputdir + "/monitoring.stats";
				// und stats dumpen
				ipv4_write_stats(filename, pcap.counters(), false);

			}

			if (trafficdump_time || hupped || !capturing)
			{
				trafficdump_time = false;

				if (!capturing)
					log2syslog("got SIGTERM signal...");

				// Pointer auf die Error-Counter struct holen
				pcap_counters *counters = pcap.counters();

				// Drops seit dem letzten Stats-Messzeitpunkt besorgen
				uint64_t deltadrops = pcap.get_delta_drops();
				// und zu den capture_drop addieren
				counters->capture_drop += deltadrops;
				counters->deltadrops = deltadrops;

				// Wenn wir noch capturen, dann ist es ein stuendlicher
				// Dump und wir muessen einen Offset von -1h verwenden, damit
				// die Files den richtigen Timestamp bekommen. Wird aber gerade
				// nicht gecaptured, dann ist entweder trafficdump_time oder hupped
				// und wir dumpen mit einem Offset von 0h
				int offset = capturing ? -1 : 0;
				if (hupped)
				{
					offset = 0;
					hupped = false;
					log2syslog("got SIGHUP signal...");
				}

				// Filename fuer Traffic-Dump zusammensetzen
				const string trafficfilename = output_filename(outputdir,
						"traffic", offset);
				// und Hash dumpen
				log2syslog("dumping to " + trafficfilename + "...");
				ipv4_write_iphash(trafficfilename);
				log2syslog("dump done");

				// wieder Drops seit dem letzten Stats-Messzeitpunkt besorgen
				deltadrops = pcap.get_delta_drops();
				// und diesmal zu den write_drop addieren
				counters->write_drop += deltadrops;

				// Filename fuer Stats-Dump zusammensetzen
				const string statsfilename = output_filename(outputdir,
						"stats", offset);
				// und stats dumpen
				ipv4_write_stats(statsfilename, pcap.counters(), true);

				// Hash leeren
				ipv4_clear_hash();

				// falls beendet werden soll, aus der while() Schleife
				// springen
				if (!capturing)
					break;
			}
		}
	}
	// exceptions von der Boost Program Options Lib
	catch (po::error &e)
	{
		ipcount_error::log2syslog(e,LOG_DAEMON, LOG_ERR);
		exit(1);
	}
	// exceptions von der c++ std lib
	catch (std::exception &e)
	{
		ipcount_error::log2syslog(e,LOG_DAEMON, LOG_ERR);
		exit(1);
	}
	// unbekannte c++ exceptions
	catch (...)
	{
		cerr << "unknown exception caught." << endl;
		exit(1);
	}

	log2syslog("ipcount exited normally.");

	exit (0);
}

